/*
 * GTK See -- a image viewer based on GTK+
 * Copyright (C) 1998 Hotaru Lee <hotaru@163.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <gtk/gtk.h>

#include "util.h"
#include "gtypes.h"
#include "scanline.h"
#include "dirtree.h"
#include "imagelist.h"
#include "gtkseemenu.h"
#include "gtkseestatus.h"
#include "gtkseeabout.h"
#include "gtkseetoolbar.h"
#include "viewer.h"
#include "gtksee.h"

#include "pixmaps/blank.xpm"
#include "pixmaps/gtksee.xpm"

#define PREVIEW_WIDTH (gdk_screen_width()-440)
#define PREVIEW_HEIGHT (gdk_screen_height()-300)

GtkWidget *entry;
GtkWidget *imagelist;
GtkWidget *preview_box;
GtkWidget *preview_image;
GtkWidget *tree, *treewin;

static GtkWidget *browser_window, *viewer_window;

static GdkPixmap *gtksee_pixmap;
static GdkBitmap *gtksee_mask;
static GdkPixmap *blank_pixmap;
static GdkBitmap *blank_mask;
static GdkCursor *watch_cursor;

static gint busy_level;
static gchar current_selected_dir[256];
static gboolean fast_preview;

int		main			(int argc,
					 char *argv[]);
void		dir_selected		(GtkTree *dt,
					 GtkWidget *child);
void		dir_selected_internal	(GtkTree *dt,
					 GtkWidget *child);
void		file_selected		(ImageList *il);
void		file_selected_internal	(ImageList *il);
void		file_unselected		(ImageList *il);
void		dir_entry_enter		(GtkWidget *widget, GtkWidget *entry);
void		set_busy_cursor		();
void		set_normal_cursor	();
void		refresh_status		();

gboolean
get_current_image(gchar *buffer)
{
	ImageInfo *info;
	
	if ((info = image_list_get_selected(IMAGE_LIST(imagelist))) == NULL)
	{
		buffer[0] = '\0';
		return FALSE;
	} else
	{
		strcpy(buffer, current_selected_dir);
		if (buffer[strlen(buffer) - 1] != '/') strcat(buffer,"/");
		strcat(buffer, info->name);
		return TRUE;
	}
}

void
refresh_list()
{
	GList *list;
	
	if (busy_level > 0) return;
	
	set_busy_cursor();
	list = gtk_container_children(GTK_CONTAINER(preview_box));
	if (list != NULL)
	{
		gtk_container_remove(
			GTK_CONTAINER(preview_box),
			GTK_WIDGET(list->data));
	}
	image_list_refresh(IMAGE_LIST(imagelist));
	refresh_status();
	set_normal_cursor();
}

void
dir_entry_enter(GtkWidget *widget, GtkWidget *entry)
{
	gchar entry_text[256];
	GtkAdjustment *vadj;
	gint pos;

	if (busy_level > 0) return;
	
	strcpy(entry_text, gtk_entry_get_text(GTK_ENTRY(entry)));
	if (strlen(entry_text) <= 0) return;
	if (entry_text[0] != '/')
	{
		strcpy(entry_text, current_selected_dir);
		if (entry_text[strlen(entry_text) - 1] != '/')
			strcat(entry_text, "/");
		strcat(entry_text, gtk_entry_get_text(GTK_ENTRY(entry)));
	}
	pos = dirtree_select_dir(GTK_TREE(tree), entry_text);
	if (pos >= 0)
	{
		vadj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(treewin));
		gtk_adjustment_set_value(
			vadj,
			min(vadj->upper - vadj->page_size,
				DIRTREE(tree)->item_height * pos
			)
		);
	}
}

void
set_busy_cursor()
{
	if (busy_level == 0)
	{
		gdk_window_set_cursor(browser_window->window, watch_cursor);
		while (gtk_events_pending()) gtk_main_iteration();
		gtk_grab_add(imagelist);
	}
	busy_level++;
}

void
set_normal_cursor()
{
	busy_level--;
	if (busy_level == 0)
	{
		gdk_window_set_cursor(browser_window->window, NULL);
		gtk_grab_remove(imagelist);
	}
}

void
dir_selected(GtkTree *dt, GtkWidget *child)
{
	if (busy_level > 0) return;
	dir_selected_internal(dt, child);
}

void
dir_selected_internal(GtkTree *dt, GtkWidget *child)
{	
	GList *list;
	
	set_busy_cursor();
	list = gtk_container_children(GTK_CONTAINER(preview_box));
	if (list != NULL)
	{
		gtk_container_remove(
			GTK_CONTAINER(preview_box),
			GTK_WIDGET(list->data));
	}

	gtk_entry_set_text(GTK_ENTRY(entry), DIRTREE(dt)->selected_dir);
	strcpy(current_selected_dir, DIRTREE(dt)->selected_dir);
	image_list_set_dir(IMAGE_LIST(imagelist), DIRTREE(dt)->selected_dir);
	
	refresh_status();
	set_normal_cursor();
}

void
file_selected(ImageList *il)
{
	if (busy_level > 0) return;
	
	file_selected_internal(il);
}

void
file_selected_internal(ImageList *il)
{
	gchar *text;
	gchar fullname[256], buffer[256];
	ImageInfo *info = image_list_get_selected(il);
	struct tm *time;
	gint hour;
	GList *list;
	gint max_width, max_height;
	gint old_width;
	
	toolbar_remove_enable(TRUE);
	toolbar_rename_enable(TRUE);
	
	max_width = preview_box->allocation.width - 4;
	max_height = preview_box->allocation.height - 4;

	/* setting status_file(size, date/time) */
	text = fsize(info->size);
	time = localtime(&info->time);
	hour = time->tm_hour % 12;
	hour = (hour==0 ? 12 : hour);
	sprintf(buffer, "%s, %02i/%02i/%02i %02i:%02i %s",
		text, time->tm_year, time->tm_mon+1, time->tm_mday,
		hour, time->tm_min, (time->tm_hour<12)?"AM":"PM"
	);
	g_free(text);
	set_status_file(buffer);
	
	/* setting status_type(pixmap) */
	set_status_type(info->type_pixmap, info->type_mask);
	
	/* setting status_name(filename) */
	if (info->has_desc)
	{
		sprintf(buffer, "%s (%s)", info->name, info->desc);
		set_status_name(buffer);
	} else
	{
		set_status_name(info->name);
	}
	
	/* setting status_prop(image properties) */
	text = g_malloc(30);
	image_convert_info(info, text);
	set_status_prop(text);

	list = gtk_container_children(GTK_CONTAINER(preview_box));
	if (list != NULL)
	{
		gtk_container_remove(
			GTK_CONTAINER(preview_box),
			GTK_WIDGET(list->data));
	}

	if (! info -> valid) return;

	/* load the image if necessary */
	if (! info -> loaded)
	{
		set_busy_cursor();
		info -> loaded = TRUE;
		strcpy(fullname, image_list_get_dir(il));
		if (fullname[strlen(fullname) - 1] != '/') strcat(fullname,"/");
		strcat(fullname, info->name);
		old_width = info->width;
		info->image = load_scaled_image(fullname, info,
			max_width, max_height,
			ROTATE_NONE,
			il -> parent,
			preview_box,
			SCANLINE_DISPLAY | (fast_preview?SCANLINE_PREVIEW:0));
		set_normal_cursor();
		if (info->width != old_width)
		{
			/* we should refresh the image info */
			image_list_update_info(il, info);
			image_convert_info(info, buffer);
			set_status_prop(buffer);
			set_status_type(info->type_pixmap, info->type_mask);
			toolbar_view_enable(image_list_get_first(IMAGE_LIST(imagelist)) != NULL);
		}
		if (info->has_desc && info->desc != NULL)
		{
			sprintf(buffer, "%s (%s)", info->name, info->desc);
			set_status_name(buffer);
			info->has_desc = TRUE;
		}
	} else
	{
		/* the image has been load before. just show it. */
		preview_image = gtk_image_new(info->image, NULL);
		gtk_widget_show(preview_image);
		gtk_container_add(GTK_CONTAINER(preview_box), preview_image);
	}
}

void
file_unselected(ImageList *il)
{
	if (busy_level > 0) return;
	
	toolbar_remove_enable(FALSE);
	toolbar_rename_enable(FALSE);
}

/* This method is called from image viewer */
void
show_browser()
{
	gtk_widget_destroy(viewer_window);
	gtk_widget_show(browser_window);
}

void
menu_file_view(GtkWidget *widget, gpointer data)
{
	if (busy_level > 0) return;
	
	if (image_list_get_first(IMAGE_LIST(imagelist)) == NULL) return;
	gtk_widget_hide(browser_window);
	viewer_window = get_viewer_window(imagelist);
}

void
menu_file_refresh(GtkWidget *widget, gpointer data)
{
	gint serial;
	ImageInfo *info;
	GtkAdjustment *vadj;
	
	if (busy_level > 0) return;
	
	set_busy_cursor();
	
	info = image_list_get_selected(IMAGE_LIST(imagelist));
	serial = (info==NULL) ? -1 : info->serial;
	
	dirtree_refresh(DIRTREE(tree));
	vadj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(treewin));
	gtk_adjustment_set_value(vadj,
		DIRTREE(tree)->item_height *
		dirtree_select_dir(GTK_TREE(tree), DIRTREE(tree)->selected_dir));
	dir_selected_internal(GTK_TREE(tree), NULL);
	
	if (serial >= 0)
	{
		image_list_select_by_serial(IMAGE_LIST(imagelist), serial);
		file_selected_internal(IMAGE_LIST(imagelist));
	}
	set_normal_cursor();
}

void
menu_file_about(GtkWidget *widget, gpointer data)
{
	GtkWidget *dialog;
	
	dialog = get_about_dialog();
	gtk_widget_show(dialog);
}

void
menu_file_quit(GtkWidget *widget, gpointer data)
{
	gtk_main_quit();
}

void
menu_view_hide(GtkWidget *widget, gpointer data)
{
	if (busy_level > 0) return;
	
	image_list_hide_non_images(
		IMAGE_LIST(imagelist),
		GTK_CHECK_MENU_ITEM(widget)->active);
	refresh_status();
}

void
menu_view_thumbnails(GtkWidget *widget, gpointer data)
{
	if (GTK_CHECK_MENU_ITEM(widget)->active)
	{
		toolbar_set_list_type(IMAGE_LIST_THUMBNAILS);
	}
}

void
menu_view_small_icons(GtkWidget *widget, gpointer data)
{
	if (GTK_CHECK_MENU_ITEM(widget)->active)
	{
		toolbar_set_list_type(IMAGE_LIST_SMALL_ICONS);
	}
}

void
menu_view_details(GtkWidget *widget, gpointer data)
{
	if (GTK_CHECK_MENU_ITEM(widget)->active)
	{
		toolbar_set_list_type(IMAGE_LIST_DETAILS);
	}
}

void
menu_view_preview(GtkWidget *widget, gpointer data)
{
	fast_preview = GTK_CHECK_MENU_ITEM(widget)->active;
	image_list_fast_preview(IMAGE_LIST(imagelist), fast_preview);
}

void
toolbar_cdup(GtkWidget *widget, gpointer data)
{
	GtkAdjustment *vadj;
	gint n;
	
	if (busy_level > 0) return;
	
	vadj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(treewin));
	n = dirtree_cdup(DIRTREE(tree));
	gtk_adjustment_set_value(vadj, DIRTREE(tree)->item_height * n);
}

void
toolbar_refresh(GtkWidget *widget, gpointer data)
{
	menu_file_refresh(widget, data);
}

void
toolbar_view(GtkWidget *widget, gpointer data)
{
	menu_file_view(widget, data);
}

void
toolbar_help(GtkWidget *widget, gpointer data)
{
	menu_file_about(widget, data);
}

void
toolbar_thumbnails(GtkWidget *widget, gpointer data)
{
	if (busy_level > 0) return;
	
	if (GTK_TOGGLE_BUTTON(widget)->active)
	{
		set_busy_cursor();
		image_list_set_list_type(IMAGE_LIST(imagelist), IMAGE_LIST_THUMBNAILS);
		refresh_status();
		menu_set_list_type(IMAGE_LIST_THUMBNAILS);
		set_normal_cursor();
	}
}

void
toolbar_small_icons(GtkWidget *widget, gpointer data)
{
	if (busy_level > 0) return;
	
	if (GTK_TOGGLE_BUTTON(widget)->active)
	{
		set_busy_cursor();
		image_list_set_list_type(IMAGE_LIST(imagelist), IMAGE_LIST_SMALL_ICONS);
		refresh_status();
		menu_set_list_type(IMAGE_LIST_SMALL_ICONS);
		set_normal_cursor();
	}
}

void
toolbar_details(GtkWidget *widget, gpointer data)
{
	if (busy_level > 0) return;
	
	if (GTK_TOGGLE_BUTTON(widget)->active)
	{
		set_busy_cursor();
		image_list_set_list_type(IMAGE_LIST(imagelist), IMAGE_LIST_DETAILS);
		refresh_status();
		menu_set_list_type(IMAGE_LIST_DETAILS);
		set_normal_cursor();
	}
}

void
refresh_status()
{
	gchar buffer[64];
	
	/* set status_main(total files and size) */
	sprintf(buffer, "Total %i file(s) (%s)",
		IMAGE_LIST(imagelist)->nfiles,
		fsize(IMAGE_LIST(imagelist)->total_size));
	set_status_main(buffer);
	
	/*
	 * When a new dir is selected, no file item are being selected.
	 * so, clear the status bar.
	 */
	set_status_file("");
	set_status_type(NULL, NULL);
	set_status_name("");
	set_status_prop("");
	toolbar_remove_enable(FALSE);
	toolbar_rename_enable(FALSE);
	toolbar_view_enable(image_list_get_first(IMAGE_LIST(imagelist)) != NULL);
}

int
main (int argc, char *argv[])
{
	GtkWidget *main_vbox;
	GtkWidget *menubar, *toolbar;
	GtkWidget *statusbar;
	GtkWidget *window;
	GtkWidget *hpaned, *vpaned, *vbox;
	GtkAdjustment *vadj;
	gchar wd[256];
	gint pos;
	
	GtkStyle *style;

	gtk_init (&argc, &argv);

	/* creating main window */
	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_signal_connect (GTK_OBJECT(window), "delete_event",
		GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
	gtk_container_border_width(GTK_CONTAINER(window), 2);

	gtk_window_set_title(GTK_WINDOW(window), "GTK See");
	gtk_window_set_wmclass(GTK_WINDOW(window), "browser", "GTKSee");
	gtk_widget_set_uposition(window, 10, 10);
	browser_window = window;
	
	gtk_widget_realize(window);

	style = gtk_widget_get_style(window);
	gtksee_pixmap = gdk_pixmap_create_from_xpm_d(
		window->window,
		&gtksee_mask, &style->bg[GTK_STATE_NORMAL],
		(gchar **)gtksee_xpm);
	gdk_window_set_icon(window->window, NULL, gtksee_pixmap, gtksee_mask);
	gdk_window_set_icon_name(window->window, "GTK See Icon");
	
	/* creating dirtree */
#ifdef HAVE_GETCWD
	getcwd(wd, 256);
#else
	strcpy(wd, "/");
#endif
	tree = dirtree_new_by_dir(window, wd);
	gtk_signal_connect(GTK_OBJECT(tree), "selection_changed", GTK_SIGNAL_FUNC(dir_selected), NULL);
	treewin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(treewin),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_widget_set_usize(treewin, PREVIEW_WIDTH, 170);
	gtk_container_add(GTK_CONTAINER(treewin), tree);
	gtk_widget_show (tree);
	gtk_widget_show(treewin);

	/* creating pixmap */
	blank_pixmap = gdk_pixmap_create_from_xpm_d(
		window->window,
		&blank_mask, &style->bg[GTK_STATE_NORMAL],
		(gchar **)blank_xpm);
	
	/* we will create this when needed. */
	preview_image = NULL;
	
	preview_box = gtk_frame_new(NULL);
	gtk_widget_set_usize(preview_box, PREVIEW_WIDTH, PREVIEW_HEIGHT);
	gtk_widget_show(preview_box);
	
	/* adding tree & pixmap into vpaned */
	vpaned = gtk_vpaned_new();
	gtk_paned_add1(GTK_PANED(vpaned), treewin);
	gtk_paned_add2(GTK_PANED(vpaned), preview_box);
	gtk_widget_show(vpaned);
	
	/* creating entry */
	entry = gtk_entry_new();
	gtk_signal_connect(
		GTK_OBJECT(entry),
		"activate",
		(GtkSignalFunc)dir_entry_enter,
		entry
	);
	gtk_widget_show(entry);
	
	/* creating imagelist */
	imagelist = image_list_new(window);
	gtk_signal_connect_object(GTK_OBJECT(imagelist),
		"select_image",
		GTK_SIGNAL_FUNC(file_selected),
		GTK_OBJECT(imagelist));
	gtk_signal_connect_object(GTK_OBJECT(imagelist),
		"unselect_image",
		GTK_SIGNAL_FUNC(file_unselected),
		GTK_OBJECT(imagelist));
	gtk_widget_set_usize(imagelist, 390, 370);
	gtk_widget_show(imagelist);
	
	/* adding entry & imagelist into vbox */
	vbox = gtk_vbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), entry, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), imagelist, TRUE, TRUE, 0);
	gtk_widget_show(vbox);
	
	/* adding vpaned * vbox into hpaned */
	hpaned = gtk_hpaned_new();
	gtk_paned_add1(GTK_PANED(hpaned), vpaned);
	gtk_paned_add2(GTK_PANED(hpaned), vbox);
	gtk_widget_show(hpaned);

	/* creating menubar */
	menubar = get_main_menu(window);
	gtk_widget_show(menubar);

	/* creating toolbar */
	toolbar = get_main_toolbar(window);
	gtk_widget_show(toolbar);
	
	/* creating statusbar */
	statusbar = get_main_status_bar(blank_pixmap, blank_mask);
	gtk_widget_show(statusbar);
	
	/* adding menubar, toolbar, hpaned & statusbar into main_vbox */
	main_vbox = gtk_vbox_new(FALSE, 2);
	gtk_box_pack_start(GTK_BOX(main_vbox), menubar, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(main_vbox), toolbar, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(main_vbox), hpaned, TRUE, TRUE, 0);
	gtk_box_pack_end(GTK_BOX(main_vbox), statusbar, FALSE, FALSE, 0);
	gtk_widget_show(main_vbox);

	gtk_container_add(GTK_CONTAINER(window), main_vbox);

	/* other values */
	scanline_init(gdk_color_context_new(
		gdk_window_get_visual(window->window),
		gdk_window_get_colormap(window->window)),
		style);
	watch_cursor = gdk_cursor_new(GDK_WATCH);
	busy_level = 1;
	toolbar_set_list_type(IMAGE_LIST_DETAILS);
	menu_set_list_type(IMAGE_LIST_DETAILS);
	busy_level = 0;
	
	gtk_widget_show(window);
	
	/* selecting current working directory */
#ifdef HAVE_GETCWD
	getcwd(wd, 256);
#else
	strcpy(wd, "/");
#endif
	pos = dirtree_select_dir(GTK_TREE(tree), wd);
	if (pos >=0) {
		vadj = gtk_scrolled_window_get_vadjustment(GTK_SCROLLED_WINDOW(treewin));
		gtk_adjustment_set_value(
			vadj,
			min(vadj->upper - vadj->page_size,
				DIRTREE(tree)->item_height * pos));
	}

	gtk_main();
	return 0;
}
