/*
 * GTK See -- a image viewer based on GTK+
 * Copyright (C) 1998 Hotaru Lee <hotaru@163.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define LABEL_HEIGHT 17
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>

#include "gtypes.h"
#include "scanline.h"
#include "imagetnitem.h"

static void	image_tnitem_class_init		(ImageTnItemClass *klass);
static void	image_tnitem_init		(ImageTnItem *ii);
static void	image_tnitem_draw		(GtkWidget *widget,
						 GdkRectangle *area);
static gint	image_tnitem_expose		(GtkWidget *widget,
						 GdkEventExpose *event);
static void	image_tnitem_paint		(GtkWidget *widget,
						 GdkRectangle *area);
static void	image_tnitem_realize		(GtkWidget *widget);
static GdkColor*
		get_tnitem_tooltips_bgcolor	(GdkWindow *window,
						 GdkColormap *colormap);

static void
image_tnitem_class_init(ImageTnItemClass *klass)
{
	GtkWidgetClass *widget_class;
	
	widget_class = (GtkWidgetClass*) klass;
	
	widget_class->draw = image_tnitem_draw;
	widget_class->realize = image_tnitem_realize;
	widget_class->expose_event = image_tnitem_expose;
}

static void
image_tnitem_init(ImageTnItem *ii)
{
	ii -> filename[0] = '\0';
	ii -> color_set = FALSE;
	ii -> image = NULL;
	ii -> bg_gc = NULL;
	ii -> preview = FALSE;
}

guint
image_tnitem_get_type()
{
	static guint ii_type = 0;

	if (!ii_type)
	{
		GtkTypeInfo ii_info =
		{
			"ImageTnItem",
			sizeof (ImageTnItem),
			sizeof (ImageTnItemClass),
			(GtkClassInitFunc) image_tnitem_class_init,
			(GtkObjectInitFunc) image_tnitem_init,
			(GtkArgSetFunc) NULL,
			(GtkArgGetFunc) NULL
		};
		ii_type = gtk_type_unique (gtk_widget_get_type(), &ii_info);
	}
	return ii_type;
}

GtkWidget*
image_tnitem_new()
{
	GtkWidget *ii;
	
	ii = GTK_WIDGET(gtk_type_new(image_tnitem_get_type()));
	gtk_widget_set_usize(ii, THUMBNAIL_WIDTH+8, THUMBNAIL_HEIGHT+12+LABEL_HEIGHT);
	return ii;
}

void
image_tnitem_set_filename(ImageTnItem *ii, guchar *filename)
{
	guchar *name;
	GtkTooltips *tooltips;
	
	strcpy(ii->filename, filename);
	name = strrchr(filename, '/');
	if (name == NULL) name = filename;
	else name++;
	tooltips = gtk_tooltips_new();
	gtk_tooltips_set_delay(tooltips, 500);
	gtk_tooltips_set_tip(tooltips, GTK_WIDGET(ii), name, NULL);
      	
	/* load the image here */
	
	gtk_widget_queue_draw(GTK_WIDGET(ii));
}

void
image_tnitem_set_color(ImageTnItem *ii, GdkColor *color)
{
	if (color == NULL)
	{
		ii->color_set = FALSE;
	} else
	{
		ii->color_set = TRUE;
		ii->color.pixel = color->pixel;
	}
	gtk_widget_queue_draw(GTK_WIDGET(ii));
}

void
image_tnitem_destroy(ImageTnItem *ii)
{
	if (ii->image != NULL) gdk_image_destroy(ii->image);
	if (ii->bg_gc != NULL) gdk_gc_destroy(ii->bg_gc);
	gtk_widget_destroy(GTK_WIDGET(ii));
}

static void
image_tnitem_draw(GtkWidget *widget, GdkRectangle *area)
{
	if (GTK_WIDGET_DRAWABLE(widget))
	{
		image_tnitem_paint(widget, area);
		gtk_widget_draw_default(widget);
		gtk_widget_draw_focus(widget);
	}
}

static void
image_tnitem_paint(GtkWidget *widget, GdkRectangle *area)
{
	ImageTnItem *item;
	guchar label_str[256], *name;
	gint i;
	
	if (GTK_WIDGET_DRAWABLE(widget))
	{
		item = IMAGE_TNITEM(widget);
		
		/* clear area */
		gtk_style_set_background(
			widget->style,
			widget->window,
			GTK_STATE_NORMAL);
		gdk_window_clear_area(widget->window,
			area->x, area->y, area->width, area->height);
		
		/* draw shadows, fill label with bg_gc if needed */
		gtk_draw_shadow(
			widget->style,
			widget->window,
			GTK_STATE_NORMAL,
			GTK_SHADOW_OUT,
			0, 0, THUMBNAIL_WIDTH+8, THUMBNAIL_HEIGHT+8);
		if (GTK_WIDGET_STATE(widget) == GTK_STATE_SELECTED)
		{
			gdk_draw_rectangle(
				widget->window,
				widget->style->bg_gc[GTK_STATE_SELECTED],
				TRUE,
				0, THUMBNAIL_HEIGHT+10,
				THUMBNAIL_WIDTH+8, LABEL_HEIGHT+2);
		} else if (item->color_set)
		{
			gdk_gc_set_foreground(item->bg_gc, &item->color);
			gdk_draw_rectangle(
				widget->window,
				item->bg_gc,
				TRUE,
				0, THUMBNAIL_HEIGHT+10,
				THUMBNAIL_WIDTH+8, LABEL_HEIGHT+2);
		}
		gtk_draw_shadow(
			widget->style,
			widget->window,
			GTK_WIDGET_STATE(widget),
			GTK_SHADOW_IN,
			0, THUMBNAIL_HEIGHT+10,
			THUMBNAIL_WIDTH+8, LABEL_HEIGHT+2);
		
		/* draw label */
		if (strlen(item->filename) < 1) return;
		name = strrchr(item->filename, '/');
		if (name == NULL) name = item->filename;
		else name++;
		strcpy(label_str, name);
		if (gdk_string_width(widget->style->font, label_str) > THUMBNAIL_WIDTH)
		{
			i = strlen(label_str);
			if (i < 3) return;
			label_str[i-1] = '\0';
			strcat(label_str, "...");
			i = strlen(label_str);
			while (i > 3 && gdk_string_width(widget->style->font, label_str) > THUMBNAIL_WIDTH)
			{
				label_str[i-1] = '\0';
				label_str[i-4] = '.';
				i--;
			}
		}
		gdk_draw_string(
			widget->window,
			widget->style->font,
			widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
			4, THUMBNAIL_HEIGHT+9+widget->style->font->ascent
			+widget->style->font->descent, label_str);
		
		/* draw the thumbnail image */
		if (item->image != NULL)
		{
			gdk_draw_image(
				widget->window,
				widget->style->fg_gc[GTK_STATE_NORMAL],
				item->image,
				0, 0,
				4+(THUMBNAIL_WIDTH-item->image->width)/2,
				4+(THUMBNAIL_HEIGHT-item->image->height)/2,
				item->image->width,
				item->image->height);
		}
	}
}

static void
image_tnitem_realize(GtkWidget *widget)
{
	ImageTnItem *item;
	GdkWindowAttr attributes;
	gint attributes_mask;
	GtkTooltipsData *tip_data;
	
	g_return_if_fail(widget != NULL);
	g_return_if_fail(IS_IMAGE_TNITEM(widget));
	
	item = IMAGE_TNITEM(widget);
	GTK_WIDGET_SET_FLAGS(widget, GTK_REALIZED);
	
	attributes.window_type = GDK_WINDOW_CHILD;
	attributes.x = widget->allocation.x;
	attributes.y = widget->allocation.y;
	attributes.width = widget->allocation.width;
	attributes.height = widget->allocation.height;
	attributes.wclass = GDK_INPUT_OUTPUT;
	attributes.visual = gtk_widget_get_visual (widget);
	attributes.colormap = gtk_widget_get_colormap (widget);
	attributes.event_mask = gtk_widget_get_events (widget);
	attributes.event_mask |= (
		GDK_EXPOSURE_MASK |
		GDK_BUTTON_PRESS_MASK |
		GDK_BUTTON_RELEASE_MASK |
		GDK_ENTER_NOTIFY_MASK |
		GDK_LEAVE_NOTIFY_MASK);
	attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
	widget->window = gdk_window_new (gtk_widget_get_parent_window (widget), &attributes, attributes_mask);
	gdk_window_set_user_data (widget->window, item);
	
	widget->style = gtk_style_attach (widget->style, widget->window);
	gtk_style_set_background(widget->style, widget->window, GTK_STATE_NORMAL);
	
	item->bg_gc = gdk_gc_new(widget->window);
	
	if ((tip_data = gtk_tooltips_data_get(widget)) != NULL &&
	    tip_data->tooltips != NULL)
	{
		gtk_tooltips_set_colors(
			tip_data->tooltips,
			get_tnitem_tooltips_bgcolor(widget->window, gtk_widget_get_colormap (widget)),
			&widget->style->fg[GTK_STATE_NORMAL]);
	}
	if (strlen(item->filename) > 0)
	{
		item->image = load_scaled_image(
			item->filename,
			(ImageInfo*) gtk_object_get_user_data(GTK_OBJECT(widget)),
			THUMBNAIL_WIDTH, THUMBNAIL_HEIGHT,
			ROTATE_NONE,
			widget,
			NULL,
			SCANLINE_INTERACT | (item->preview?SCANLINE_PREVIEW:0));
	}
}

static gint
image_tnitem_expose(GtkWidget *widget, GdkEventExpose *event)
{
	ImageTnItem *item;
	
	g_return_val_if_fail(widget != NULL, FALSE);
	g_return_val_if_fail(IS_IMAGE_TNITEM(widget), FALSE);
	g_return_val_if_fail(event != NULL, FALSE);
	
	if (GTK_WIDGET_DRAWABLE(widget))
	{
		item = IMAGE_TNITEM(widget);
		image_tnitem_paint(widget, &event->area);
		gtk_widget_draw_default(widget);
		gtk_widget_draw_focus(widget);
	}
	return FALSE;
}

static GdkColor*
get_tnitem_tooltips_bgcolor(GdkWindow *window, GdkColormap *colormap)
{
	static GdkColor color;
	static gboolean alloced = FALSE;
	if (!alloced)
	{
		color.red = 61669;
		color.green = 59113;
		color.blue = 35979;
		color.pixel = 0;
		gdk_color_alloc(colormap, &color);
		alloced = TRUE;
	}
	return &color;
}

void
image_tnitem_set_preview(ImageTnItem *ii, gboolean preview)
{
	ii->preview = preview;
}

